// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO64C/16AO64C_Linux_1.x.x.x.GSC_DN/driver/device.c $
// $Rev: 53038 $
// $Date: 2023-05-18 15:57:30 -0500 (Thu, 18 May 2023) $

// 16AO64C: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#ifndef	DEV_DEVICE_CREATE_EXTERN
#define	DEV_DEVICE_CREATE_EXTERN(dev)		0
#endif



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type
	{ "16AO64C",	0x10B5, 0x9056, 0x10B5, 0x3555,	GSC_DEV_TYPE_16AO64C	},
	{ NULL }
};



//*****************************************************************************
static void _autocal_ms_compute(dev_data_t* dev)
{
	if (dev->cache.channel_qty <= 16)
		dev->cache.autocal_ms	= 5000;
	else if (dev->cache.channel_qty <= 32)
		dev->cache.autocal_ms	= 7000;
	else
		dev->cache.autocal_ms	= 11000;
}



//*****************************************************************************
static void _channels_qty_compute(dev_data_t* dev)
{
	int	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 17, 16);

	dev->cache.channels_max	= 64;

	switch (field)
	{
		default:
		case 0:		dev->cache.channel_qty	= 64;
					break;

		case 1:
		case 2:		dev->cache.channel_qty	= 32;
					break;

		case 3:		dev->cache.channel_qty	= 16;
					break;
	}
}



//*****************************************************************************
static void _channels_type_compute(dev_data_t* dev)
{
	int	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 17, 16);

	switch (field)
	{
		default:
		case 0:
		case 1:		dev->cache.differential	= 0;
					break;

		case 2:
		case 3:		dev->cache.differential	= 1;
					break;
	}
}



//*****************************************************************************
static void _fref_default_compute(dev_data_t* dev)
{
	int	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 23, 22);

	switch (field)
	{
		default:
		case 0:		dev->cache.fref_default	= 49152000L;
					dev->cache.nrate_mask	= 0xFFFFFF;
					dev->cache.nrate_max	= 0xFFFFFF;
					dev->cache.nrate_min	= 98;
					dev->cache.fsamp_max	= 501551;
					dev->cache.fsamp_min	= 3;
					break;
	}
}



//*****************************************************************************
static void _filter_compute(dev_data_t* dev)
{
	int	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 21, 20);

	switch (field)
	{
		case 0:		dev->cache.filter	= AO64C_FILTER_100KHZ;
					break;

		case 1:		dev->cache.filter	= AO64C_FILTER_10KHZ;
					break;

		default:
		case 2:		dev->cache.filter	= AO64C_FILTER_NONE;
					break;
	}
}



//*****************************************************************************
static void _range_compute(dev_data_t* dev)
{
	int	field	= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 19, 18);

	switch (field)
	{
		default:
		case 0:		dev->cache.range	= AO64C_RANGE_UNI_10_5;
					break;

		case 1:		dev->cache.range	= AO64C_RANGE_BI_10_5;
					break;
	}
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AO64C_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_acr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_ACR));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_BCR));
		dev->vaddr.gsc_bor_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_BOR));
		dev->vaddr.gsc_bsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_BSR));
		dev->vaddr.gsc_btr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_BTR));
		dev->vaddr.gsc_odbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_ODBR));
		dev->vaddr.gsc_ragr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_RAGR));
		dev->vaddr.gsc_rbgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO64C_GSC_RBGR));

		// Data cache initialization.
		dev->cache.gsc_acr_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acr_32);

		dev->cache.disconnect		= GSC_FIELD_DECODE(dev->cache.gsc_acr_32, 24, 24);
		dev->cache.fifo_size		= _256K;
		dev->cache.fw_version		= dev->cache.gsc_acr_32 & 0xFFFF;
		dev->cache.initialize_ms	= 3;

		_channels_qty_compute(dev);
		_channels_type_compute(dev);
		_filter_compute(dev);
		_fref_default_compute(dev);
		_range_compute(dev);
		_autocal_ms_compute(dev);

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_WRITE
			| GSC_DMA_CAP_DMDMA_WRITE;
		ret	= gsc_dma_create(dev, dma, dma);
		if (ret)	break;

		ret	= DEV_DEVICE_CREATE_EXTERN(dev);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


