// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO64C/16AO64C_Linux_1.x.x.x.GSC_DN/driver/write.c $
// $Rev: 51090 $
// $Date: 2022-05-31 10:48:15 -0500 (Tue, 31 May 2022) $

// 16AO64C: Device Driver: source file

#include "main.h"




//*****************************************************************************
static int _byte_swap(const dev_data_t* dev, u32* src, size_t bytes)
{
	#define	_LOW_		0x01
	#define	_HIGH_		0x02
	#define	_BOTH_		(_LOW_ | _HIGH_)

	int	count;
	s32	got;
	int	ret	= 0;
	s32	want;

	for (;;)	// A convenience loop.
	{
		if (dev->io.tx.io_byte_swap == AO64C_TX_IO_BYTE_SWAP_DISABLE)
			break;

		want	= (dev->cache.fw_version >= 0x400) ? _HIGH_ : _LOW_;
		got		= 0;

		// Scan the data to find a sample with a non-zero control byte.

		for (count = bytes / 4; count; count--, src++)
		{
			if (src[0] & 0xFF000000)
				got	|= _HIGH_;

			if (src[0] & 0x00FF0000)
				got	|= _LOW_;

			if (got)
				break;
		}

		if ((got == 0) || (got == want))
		{
			// Either all control byte data is clear, which is possible,
			// or the data is formatted as required.
			break;
		}

		if (dev->io.tx.io_byte_swap == AO64C_TX_IO_BYTE_SWAP_CHECK)
		{
			// We're only checking the data and it isn't formatted as required.
			ret	= -EINVAL;
			break;
		}

		// Perform the byte swapping.

		switch (got)
		{
			default:

				printk("%s: INTERNAL ERROR: %s, line %d\n",
						DEV_NAME,
						__FILE__,
						__LINE__);
				ret	= -EINVAL;
				break;

			case _BOTH_:

				// This is an unlikely occurance. If the data stream is unintentionally
				// of this format, it may be unlikely that we catch it since we only
				// look at the first sample with a non-zero upper word.
				ret	= -EINVAL;
				break;

			case _LOW_:	// want HIGH, move D23-D16 to D31-D24.

				for (count = bytes / 4; count; count--, src++)
					src[0]	= (src[0] & 0xFFFF) | ((src[0] & 0x00FF0000) << 8);

				break;

			case _HIGH_: // want LOW, move D31-D24 to D23-D16.

				for (count = bytes / 4; count; count--, src++)
					src[0]	= (src[0] & 0xFFFF) | ((src[0] & 0xFF000000) >> 8);

				break;
		}
	}

	return(ret);
}



//*****************************************************************************
static void _dev_io_sw_init(dev_data_t* dev, dev_io_t* io)
{
	io->io_mode			= AO64C_IO_MODE_DEFAULT;
	io->overflow_data	= AO64C_TX_IO_OVER_DATA_DEFAULT;
	io->overflow_frame	= AO64C_TX_IO_OVER_FRAME_DEFAULT;
	io->pio_threshold	= 32;
	io->timeout_s		= AO64C_IO_TIMEOUT_DEFAULT;
}



//*****************************************************************************
static void _dev_io_close(dev_data_t* dev, dev_io_t* io)
{
	io->dma_channel	= NULL;
}



//*****************************************************************************
static void _dev_io_open(dev_data_t* dev, dev_io_t* io)
{
	_dev_io_sw_init(dev, io);
}



//*****************************************************************************
static int _dev_io_startup(dev_data_t* dev, dev_io_t* io)
{
	u32	bor;
	int	ret	= 0;

	if ((io->overflow_data) || (io->overflow_frame))
	{
		bor	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bor_32);

		if ((io->overflow_data) && (bor & D16))
			ret	= -EIO;

		if ((io->overflow_frame) && (bor & D17))
			ret	= -EIO;
	}

	return(ret);
}



//*****************************************************************************
static long _dev_pio_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
#ifdef PIO_WRITE_AVAILABLE_EXTERN

	bytes	= PIO_WRITE_AVAILABLE_EXTERN(dev, bytes);

#else

	u32		bsr;
	s32		level;
	s32		samples;

	bsr		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bsr_32);
	level	= bsr & 0x7FFFF;

	if (level >= dev->cache.fifo_size)
		samples	= 0;
	else
		samples	= dev->cache.fifo_size - level;

	bytes	= samples * 4;

#endif

	return(bytes);
}



//*****************************************************************************
static long _dev_pio_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;
	int		ret;

	ret	= _byte_swap(dev, mem->ptr, bytes);

	if (ret)
		qty	= ret;
	else
		qty	= gsc_write_pio_work_32_bit(dev, io, mem, bytes, st_end);

	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
#ifdef DMA_WRITE_AVAILABLE_EXTERN

	bytes	= DMA_WRITE_AVAILABLE_EXTERN(dev, bytes);

#else

	bytes	= _dev_pio_available(dev, io, bytes);

#endif
	return(bytes);
}



//*****************************************************************************
static long _dev_bmdma_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty;
	int				ret;
	long			samples	= bytes / 4;
	gsc_dma_setup_t	setup;

	ret	= _byte_swap(dev, mem->ptr, bytes);

	if (ret)
	{
		qty	= ret;
	}
	else if (samples < io->pio_threshold)
	{
		qty	= gsc_write_pio_work_32_bit(dev, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= dev;
		setup.dev		= dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_BMDMA_WRITE;

		setup.mode		= GSC_DMA_MODE_BLOCK_DMA
						| GSC_DMA_MODE_SIZE_32_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_HOST_TO_BOARD
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
#ifdef DMDMA_WRITE_AVAILABLE_EXTERN

	bytes	= DMDMA_WRITE_AVAILABLE_EXTERN(dev, bytes);

#endif

	return(bytes);
}



//*****************************************************************************
static long _dev_dmdma_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty;
	int				ret;
	long			samples	= bytes / 4;
	gsc_dma_setup_t	setup;

	ret	= _byte_swap(dev, mem->ptr, bytes);

	if (ret)
	{
		qty	= ret;
	}
	else if (samples < io->pio_threshold)
	{
		qty	= gsc_write_pio_work_32_bit(dev, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= dev;
		setup.dev		= dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_DMDMA_WRITE;

		setup.mode		= GSC_DMA_MODE_DM_DMA
						| GSC_DMA_MODE_SIZE_32_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_HOST_TO_BOARD
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	dev_write_create
*
*	Purpose:
*
*		Perform a one-tine setup of the Write Analog Output streaming I/O structure.
*
*	Arguments:
*
*		dev		The data for the device of interest.
*
*		io		The I/O structure for this stream.
*
*	Returned:
*
*		None.
*
******************************************************************************/

int dev_write_create(dev_data_t* dev, dev_io_t* io)
{
	int	ret;

	io->bytes_per_sample	= 4;
	io->io_reg_offset		= GSC_REG_OFFSET(AO64C_GSC_ODBR);
	io->io_reg_vaddr		= dev->vaddr.gsc_odbr_32;

	io->dev_io_sw_init		= _dev_io_sw_init;
	io->dev_io_close		= _dev_io_close;
	io->dev_io_open			= _dev_io_open;
	io->dev_io_startup		= _dev_io_startup;
	io->dev_pio_available	= _dev_pio_available;
	io->dev_bmdma_available	= _dev_bmdma_available;
	io->dev_dmdma_available	= _dev_dmdma_available;
	io->dev_pio_xfer		= _dev_pio_xfer;
	io->dev_bmdma_xfer		= _dev_bmdma_xfer;
	io->dev_dmdma_xfer		= _dev_dmdma_xfer;

	io->wait.abort			= AO64C_WAIT_IO_TX_ABORT;
	io->wait.done			= AO64C_WAIT_IO_TX_DONE;
	io->wait.error			= AO64C_WAIT_IO_TX_ERROR;
	io->wait.timeout		= AO64C_WAIT_IO_TX_TIMEOUT;

	ret	= gsc_io_create(dev, io, dev->cache.fifo_size * 4);

	return(ret);
}


