// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO64C/16AO64C_Linux_1.x.x.x.GSC_DN/samples/aout/perform.c $
// $Rev: 53053 $
// $Date: 2023-05-19 11:07:19 -0500 (Fri, 19 May 2023) $

// 16AO64C: Sample Application: source file

#include "main.h"



// variables ******************************************************************

static	u32		_data[128 * 1024];
static	long	_loops;		// Number of times to send a block of data.
static	long	_qty;		// Number of samples to send per block.



//*****************************************************************************
static int _init_data(const args_t* args, s32 chans)
{
	#define	PI		(3.14159265359)

	u32			acr;
	int			chan;		// 0, 1, 2, 4
	u32*		dest;
	int			errs;
	u32			gef;		// Group End Flag
	int			i;
	long		period_len;	// The length of the pattern in samples.
	long		period_tot;	// Pattern samples total for all channels.
	long		period;		// Pattern length in x units.
	int			samp;		// 0, 1, 2, 3
	long		samp_dat;	// Total sample size of the data buffer.
	float		samp_sec;	// Samples Per Second for all active channels
	float		samp_tot;	// Samples Total for all active channels
	int			shift;		// For placement of control byte.
	const char*	units;

	gsc_label("Initializing Data");

	// See where the control byte goes.
	errs	= ao64c_reg_read(args->fd, -1, 0, AO64C_GSC_ACR, &acr);
	shift	= ((acr & 0xFFFF) >= 0x0400) ? 24 : 16;
	gef		= ((acr & 0xFFFF) >= 0x0400) ? 0x40000000L : 00400000L;

	memset(_data, 0, sizeof(_data));
	samp_sec	= ((float) args->fsamp) * chans;
	samp_tot	= ((float) args->fsamp) * chans * args->seconds;
	samp_dat	= sizeof(_data) / 4;

	// Assign a default pattern length in samples.
	period_len	= (args->fsamp <= 100)		? args->fsamp			//      10-      100 ->    10-   100
				: (args->fsamp <= 1000)		? (args->fsamp / 2)		//     101-    1,000 ->    50-   500
				: (args->fsamp <= 10000)	? (args->fsamp / 5)		//   1,001-   10,000 ->   200- 2,000
				: (args->fsamp <= 100000)	? (args->fsamp / 10)	//  10,001-  100,000 -> 1,000-10,000
				: (args->fsamp / 100);								// 100,001-1,000,000 -> 1,000-10,000
	period_len	= (period_len <= 1) ? 2 : period_len;
	period_tot	= period_len * chans;

	// Compute the period in ns/us/ms
	period		= (long) (1000000000. / args->fsamp * period_len);
	units		= "ns";

	if ((period % 1000) == 0)
	{
		period	/= 1000;
		units	= "us";
	}

	if ((period % 1000) == 0)
	{
		period	/= 1000;
		units	= "ms";
	}

	if (samp_tot <= SIZEOF_ARRAY(_data))
	{
		// We can get the entire image in the buffer.
		_loops	= 1;
		_qty	= (long) samp_tot;
	}
	else if (samp_sec <= SIZEOF_ARRAY(_data))
	{
		// We can get one second of the image in the buffer.
		_loops	= args->seconds;
		_qty	= (long) samp_sec;
	}
	else if (period_tot <= SIZEOF_ARRAY(_data))
	{
		// We can get one period of the image in the buffer.
		_loops	= (long) (args->seconds * (samp_sec / period_tot));
		_qty	= period_tot;
	}
	else
	{
		// We can get only a fraction of a period of the image in the buffer.
		_qty	= (long) ((samp_dat / chans) * chans);
		_loops	= (long) ((samp_tot + _qty - 1) / _qty);
	}

	// Initialize the data buffer.
	// The data is encoded in OFFSET BINARY format.
	dest	= _data;

	for (i = 0, samp = 0; i < _qty; i += chans, samp++)
	{
		chan	= 0;
		samp	= (samp < period_len) ? samp : 0;

		// 1st channel: sawtooth wave
		dest[0]	= (u16) (0xFFFF - (0xFFFF * samp / (period_len - 1)));
		// Channel tag is zero.
		dest++;
		chan++;

		// 2nd channel: ramp wave

		if (chans >= 2)
		{
			dest[0]	= (u16) (0xFFFF * samp / (period_len - 1));
			dest[0]	|= chan << shift;	// Channel Tag
			dest++;
			chan++;
		}

		// 3rd channel: square wave

		if (chans >= 3)
		{
			dest[0]	= (u16) ((samp < (period_len / 2)) ? 0xFFFF : 0);
			dest[0]	|= chan << shift;	// Channel Tag
			dest++;
			chan++;
		}

		// 4th channel: sine wave

		if (chans >= 4)
		{
			dest[0]	= (u16) ((long) 0x7FFF) * (1 + sin((2 * PI / period_len) * samp));
			dest[0]	|= chan << shift;	// Channel Tag
			dest++;
		}

		dest[-1]	|= gef;	// Group End Flag
	}

	if (errs)
	{
		printf("FAIL <---\n");
	}
	else
	{
		printf("DONE  (pattern length: ");
		gsc_label_long_comma(period_len);
		printf(" samples, ");
		gsc_label_long_comma(period);
		printf(" %s)\n", units);
	}

	return(errs);
}



//*****************************************************************************
static int _io_write(const args_t* args, int errs)
{
	os_time_ns_t	begin;
	os_time_ns_t	end;
	s32				fifo;
	long			loop;
	long long		ms;
	char*			ptr;
	double			rate;
	int				send;
	int				sent;
	s32				timeout;
	long long		total;

	for (;;)	// A convenience loop.
	{
		if (errs)
		{
			errs	= 0;
			gsc_label("Writing Data");
			printf("SKIPPED  (due to errors)\n");
			break;
		}

		// Prefill the output buffer
		gsc_label("Prefill Buffer");
		errs	+= ao64c_tx_io_timeout(args->fd, -1, 0, -1, &timeout);
		errs	+= ao64c_tx_io_timeout(args->fd, -1, 0, 0, NULL);
		errs	+= ao64c_query(args->fd, -1, 0, AO64C_QUERY_FIFO_SIZE, &fifo);
		fifo	*= 4;
		send	= _qty * 4;
		total	= 0;

		for (;;)
		{
			sent	= ao64c_write(args->fd, _data, send);

			if (sent < 0)
				errs++;

			if (errs)
				break;

			if (sent < (send / 10))
				break;

			total	+= sent;

			if (total >= (fifo * 2))	// make extra large, fsamp may be very high
				break;
		}

		errs	+= ao64c_tx_io_timeout(args->fd, -1, 0, timeout, NULL);
		printf("%s\n", errs ? "FAIL <---" : "PASS");

		if (errs)
			break;

		// Write the data to the device.
		gsc_label("Writing Data");
		total	= 0;
		os_time_get_ns(&begin);

		for (loop = 0; loop < _loops; loop++)
		{
			if (errs)
				break;

			send	= _qty * 4;
			ptr		= (char*) _data;

			for (;;)
			{
				sent	= ao64c_write(args->fd, ptr, send);

				if (sent < 0)
					errs++;

				if (sent <= 0)
					break;

				total	+= sent;
				ptr		+= sent;
				send	-= sent;

				if (send <= 0)
					break;
			}

			if (sent <= 0)
				break;
		}

		// Compute the results.
		os_time_get_ns(&end);

		if (errs)
		{
			errs	= 1;
			printf("FAIL <---  (write error, ");
		}
		else if (sent == 0)
		{
			errs	= 1;
			printf("FAIL <---  (I/O timeout, ");
		}
		else
		{
			printf("PASS  (");
		}

		// Samples
		gsc_label_long_comma(total / 4);
		printf(" Samples, ");

		// Period
		ms	= ((long long) end.tv_sec * 1000 + end.tv_nsec / 1000000)
			- ((long long) begin.tv_sec * 1000 + begin.tv_nsec / 1000000);
		gsc_label_float_comma((double) ms / 1000, -1, 3);
		printf(" Seconds, ");

		// Rate
		ms		= (ms <= 0) ? 1 : ms;
		rate	= ((double) total * 1000000 / 4) / ms / 1000;
		gsc_label_float_comma(rate, -1, 3);
		printf(" S/S)\n");
		break;
	}

	return(errs);
}



//*****************************************************************************
int perform_tests(const args_t* args)
{
	s32	chans	= 4;
	int	errs	= 0;

	errs	+= ao64c_config_ao			(args->fd, -1, 1, args->fsamp, 0xF);
	errs	+= ao64c_tx_io_mode			(args->fd, -1, 1, args->io_mode, NULL);
	errs	+= ao64c_fsamp_ao_report_all(args->fd, -1, 1, 0xF, NULL);

	gsc_label("Analog Output");
	printf("\n");
	gsc_label_level_inc();

	errs	+= _init_data(args, chans);
	errs	+= ao64c_buffer_over_frame	(args->fd, -1, 1, AO64C_BUFFER_OVER_FRAME_CLR, NULL);
	errs	+= ao64c_buffer_over_data	(args->fd, -1, 1, AO64C_BUFFER_OVER_DATA_CLR, NULL);
	errs	+= _io_write(args, errs);
	errs	+= ao64c_buffer_over_frame	(args->fd, -1, 1, AO64C_BUFFER_OVER_FRAME_CHK, NULL);
	errs	+= ao64c_buffer_over_data	(args->fd, -1, 1, AO64C_BUFFER_OVER_DATA_CHK, NULL);

	gsc_label_level_dec();
	return(errs);
}


