// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO64C/16AO64C_Linux_1.x.x.x.GSC_DN/samples/mbsync/perform.c $
// $Rev: 51650 $
// $Date: 2022-10-07 09:20:51 -0500 (Fri, 07 Oct 2022) $

// 16AO64C: Sample Application: source file

#include "main.h"



// macros *********************************************************************

#define	D30				(0x1 << 30)

#define	SCANS			(256 * 1024 / 4 / 64)	// FIFO Size / 4 / 64 channels

#define	PATTERN_LENGTH	(SCANS / 10)

#define	_1K				(1024)
#define	_1M				(_1K * _1K)
#define	_1G				(_1K * _1K * _1K)



// variables ******************************************************************

// The output data is defined to fill 1/4th of the device's output FIFO for a 64
// channel device. 32 channel devices need only half that amount of data. The
// data size is defined such that no output FIFO will run empty while another
// write request is blocked waiting for the write to complete. This can be done
// using data sets over 1/4 of the FIFO size, but I chose 1/4 for mere ease.
static	u32	_32[SCANS * 32];
static	u32	_64[SCANS * 64];



//*****************************************************************************
static void _init_data_buffer(int chans, u32* dst)
{
	int	chan;
	u32	data;
	u32	last	= D30;
	int	samp	= 0;
	int	scan;
	u32	tag;
	u32	val;

	for (scan = 0; scan < SCANS; scan++)
	{
		// Output a saw tooth wave.
		val		= (u16) (0xFFFF - (0xFFFF * samp / (PATTERN_LENGTH - 1)));
		samp	= (samp + 1) % PATTERN_LENGTH;

		for (chan = 0; chan < chans; chan++, dst++)
		{
			tag		= chan << 24;
			data	= val | tag;

			if (chan == (chans - 1))
				data	|= last;

			dst[0]	= data;
		}
	}
}



//*****************************************************************************
static int _data_init(void)
{

	gsc_label("Initializing Data");
	_init_data_buffer(32, _32);
	_init_data_buffer(64, _64);
	printf("Done\n");
	return(0);
}


//*****************************************************************************
static int _config_sync(const args_t* args)
{
	char	buf[64];
	int		errs	= 0;
	int		i;

	gsc_label("Synchronize Devices");
	printf("\n");
	gsc_label_level_inc();

	for (i = 0; (i < BOARD_QTY_MAX) && (args->index[i] >= 0); i++)
	{
		sprintf(buf, "Device #%d, ", args->index[i]);

		if (i == 0)
			strcat(buf, "Initiator");
		else
			sprintf(buf + strlen(buf), "Target #%d", i);

		gsc_label(buf);

		if (args->unsync)
			printf("UNSYNCHRONIZED CONFIGURATION");

		printf("\n");

		gsc_label_level_inc();

		errs	+= ao64c_clock_enable		(args->fd[i], args->index[i], 1, AO64C_CLOCK_ENABLE_NO,			NULL);
		errs	+= ao64c_buffer_clear		(args->fd[i], args->index[i], 1);
		errs	+= ao64c_buffer_over_data	(args->fd[i], args->index[i], 1, AO64C_BUFFER_OVER_DATA_CLR,	NULL);
		errs	+= ao64c_buffer_over_frame	(args->fd[i], args->index[i], 1, AO64C_BUFFER_OVER_FRAME_CLR,	NULL);

		if (args->unsync)
		{
			errs	+= ao64c_clock_mode		(args->fd[i], args->index[i], 1, AO64C_CLOCK_MODE_TARGET,		NULL);
			errs	+= ao64c_trigger_mode	(args->fd[i], args->index[i], 1, AO64C_TRIGGER_MODE_TARGET,		NULL);
			errs	+= ao64c_clock_src		(args->fd[i], args->index[i], 1, AO64C_CLOCK_SRC_RAG,			NULL);
			errs	+= ao64c_burst_trig_src	(args->fd[i], args->index[i], 1, AO64C_BURST_TRIG_SRC_RBG,		NULL);
		}
		else if (i == 0)
		{
			// The initiator.
			errs	+= ao64c_clock_mode		(args->fd[i], args->index[i], 1, AO64C_CLOCK_MODE_INITIATOR,	NULL);
			errs	+= ao64c_trigger_mode	(args->fd[i], args->index[i], 1, AO64C_TRIGGER_MODE_INITIATOR,	NULL);
			errs	+= ao64c_clock_src		(args->fd[i], args->index[i], 1, AO64C_CLOCK_SRC_RAG,			NULL);
			errs	+= ao64c_burst_trig_src	(args->fd[i], args->index[i], 1, AO64C_BURST_TRIG_SRC_RBG,		NULL);
		}
		else
		{
			// The targets.
			errs	+= ao64c_clock_mode		(args->fd[i], args->index[i], 1, AO64C_CLOCK_MODE_TARGET,		NULL);
			errs	+= ao64c_trigger_mode	(args->fd[i], args->index[i], 1, AO64C_TRIGGER_MODE_TARGET,		NULL);
			errs	+= ao64c_clock_src		(args->fd[i], args->index[i], 1, AO64C_CLOCK_SRC_EXT,			NULL);
			errs	+= ao64c_burst_trig_src	(args->fd[i], args->index[i], 1, AO64C_BURST_TRIG_SRC_EXT,		NULL);
		}

		gsc_label_level_dec();
	}

	gsc_label_level_dec();
	return(errs);
}


//*****************************************************************************
static int _prefill_fifos(const args_t* args, const s32* chans)
{
	char		buf[64];
	int			bytes;
	int			errs	= 0;
	int			i;
	int			qty;
	const void*	src;

	gsc_label("Prefill FIFOs");
	printf("\n");
	gsc_label_level_inc();

	for (i = 0; (i < BOARD_QTY_MAX) && (args->index[i] >= 0); i++)
	{
		sprintf(buf, "Device #%d, ", args->index[i]);

		if (i == 0)
			strcat(buf, "Initiator");
		else
			sprintf(buf + strlen(buf), "Target #%d", i);

		gsc_label(buf);

		if (chans[i] == 32)
		{
			src		= _32;
			bytes	= sizeof(_32);
		}
		else
		{
			src		= _64;
			bytes	= sizeof(_64);
		}

		// The data sources are at most 1/4 of the device's FIFO.
		qty	= 0;
		qty	+= ao64c_write_util(args->fd[i], -1, 0, src, bytes);
		qty	+= ao64c_write_util(args->fd[i], -1, 0, src, bytes);
		qty	+= ao64c_write_util(args->fd[i], -1, 0, src, bytes);

		if (qty != (3 * bytes))
		{
			errs	= 1;
			printf("FAIL <---\n");
		}
		else
		{
			printf("PASS\n");
		}
	}

	gsc_label_level_dec();
	return(errs);
}



//*****************************************************************************
static int _enable_clocking(const args_t* args)
{
	char	buf[64];
	int		err;
	int		errs	= 0;
	int		i;

	gsc_label("Enable Clocking");
	printf("\n");
	gsc_label_level_inc();

	// Enable output.

	for (i = 1; (i < BOARD_QTY_MAX) && (args->index[i] >= 0); i++)
	{
		sprintf(buf, "Clock Enable #%d, Target %d", args->index[i], i);
		gsc_label(buf);
		err		= ao64c_clock_enable(args->fd[i], -1, 0, AO64C_CLOCK_ENABLE_YES, NULL);
		errs	+= err;
		printf("%s  (Yes)\n", err ? "FAIL <---" : "PASS");
	}

	sprintf(buf, "Clock Enable #%d, Initiator", args->index[0]);
	gsc_label(buf);
	err		= ao64c_clock_enable(args->fd[0], -1, 0, AO64C_CLOCK_ENABLE_YES, NULL);
	errs	+= err;
	printf("%s  (Yes)\n", err ? "FAIL <---" : "PASS");

	gsc_label_level_dec();
	return(errs);
}



//*****************************************************************************
static void _show_bytes(long long bytes)
{
	const char*		plus;
	long long		tmp;

	gsc_label_long_comma(bytes);
	printf(" Bytes");

	plus	= ", ";

	if (bytes > _1G)
	{
		tmp		= bytes / _1G;
		bytes	-= tmp * _1G;
		printf("%s", plus);
		plus	= " + ";
		gsc_label_long_comma(tmp);
		printf("G");
	}

	if (bytes > _1M)
	{
		tmp		= bytes / _1M;
		bytes	-= tmp * _1M;
		printf("%s", plus);
		plus	= " + ";
		gsc_label_long_comma(tmp);
		printf("M");
	}

	if (bytes > _1K)
	{
		tmp		= bytes / _1K;
		bytes	-= tmp * _1K;
		printf("%s", plus);
		plus	= " + ";
		gsc_label_long_comma(tmp);
		printf("K");
	}

	if (bytes > 0)
	{
		printf("%s", plus);
		gsc_label_long_comma(bytes);
	}
}



//*****************************************************************************
static int _output_data(const args_t* args, const s32* chans, int errs)
{
	os_time_ns_t	begin;
	int				bytes;
	int				i;
	long long		ms;
	long long		ms_limit	= (long long) args->seconds * 1000;
	os_time_ns_t	now;
	int				qty;
	const void*		src;
	long long		total	= 0;

	gsc_label("Output Data");
	gsc_label_long_comma(args->seconds);
	printf(" Seconds\n");
	gsc_label_level_inc();

	gsc_label("Working");

	if (errs)
	{
		errs	= 0;
		printf("ABORTING DUE TO ERRORS\n");
	}
	else
	{
		fflush(stdout);
		os_time_get_ns(&begin);

		for (; errs == 0;)
		{
			for (i = 1; (i <= BOARD_QTY_MAX) && (args->index[i] >= 0); i++)
			{
				if (i == BOARD_QTY_MAX)		// We want the INITIATOR done last.
					i = 0;

				if (chans[i] == 32)
				{
					src		= _32;
					bytes	= sizeof(_32);
				}
				else
				{
					src		= _64;
					bytes	= sizeof(_64);
				}

				// The data source sizes are at most 1/4 of the device's FIFO.
				qty	= ao64c_write_util(args->fd[i], -1, 0, src, bytes);

				if (qty < 0)
				{
					errs	= 1;
					printf("FAIL <---  (write error: %d\n", qty);
					break;
				}
				else if (qty != bytes)
				{
					errs	= 1;
					printf(	"FAIL <---  (write error: requested %d, wrote %d)\n",
							bytes,
							qty);
					break;
				}
				else
				{
					total	+= qty;
				}

				if (i == 0)		// We want the INITIATOR done last.
					break;
			}

			os_time_get_ns(&now);

			ms	= ((long long) now.tv_sec * 1000 + now.tv_nsec / 1000000)
				- ((long long) begin.tv_sec * 1000 + begin.tv_nsec / 1000000);

			if (ms >= ms_limit)
				break;
		}

		if (errs)
			gsc_label(NULL);

		printf("%s  (wrote ", errs ? "FAIL <---" : "PASS");
		_show_bytes(total);
		printf(")\n");
	}

	gsc_label_level_dec();
	return(errs);
}



//*****************************************************************************
int perform_tests(const args_t* args)
{
	s32	chans[BOARD_QTY_MAX];	// mirrors index and fd lists
	int	errs	= 0;
	s32	fsamp	= 5000;

	memset(chans, 0, sizeof(chans));

	// These functions are in service.c
	errs	+= service_config_ao(args, chans, fsamp);
	errs	+= service_query_channels(args, chans);
	errs	+= service_sample_rate(args, chans);

	// These functions are in this source file.
	errs	+= _data_init();
	errs	+= _config_sync(args);
	errs	+= _prefill_fifos(args, chans);
	errs	+= _enable_clocking(args);
	errs	+= _output_data(args, chans, errs);

	return(errs);
}


