// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO64C/16AO64C_Linux_1.x.x.x.GSC_DN/utils/util_fsamp_ao.c $
// $Rev: 53036 $
// $Date: 2023-05-18 15:52:40 -0500 (Thu, 18 May 2023) $

// 16AO64C: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



/******************************************************************************
*
*	Function:	ao64c_fsamp_ao_compute
*
*	Purpose:
*
*		Determine the best configuration for the specified sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	0: errors only, !0: all output
*
*		fsamp	The desired sample rate.
*
*		nrate	The Nrate value is put here. MUST NOT BE NULL!
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao64c_fsamp_ao_compute(int fd, int index, int verbose, s32 fsamp, s32* nrate, double* rate)
{
	double	delta;
	double	delta_best	= FLT_MAX;
	int		errs		= 0;
	s32		fref;
	double	fsamp_best	= 0;
	double	fsamp_got;
	s32		fsamp_max;	// Queried value.
	s32		fsamp_min;	// Queried value.
	double	fsamp_want;
	s32		nrate_best	= 0;
	s32		nrate_ll;	// lower limit
	s32		nrate_max;	// Queried value.
	s32		nrate_min;	// Queried value.
	s32		nrate_ul;	// upper limit
	s32		nrate_use;

	if (verbose)
	{
		gsc_label_index("AO Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	errs	+= ao64c_query(fd, -1, 0, AO64C_QUERY_FREF_DEFAULT, &fref);
	errs	+= ao64c_query(fd, -1, 0, AO64C_QUERY_FSAMP_MAX, &fsamp_max);
	errs	+= ao64c_query(fd, -1, 0, AO64C_QUERY_FSAMP_MIN, &fsamp_min);
	errs	+= ao64c_query(fd, -1, 0, AO64C_QUERY_NRATE_MAX, &nrate_max);
	errs	+= ao64c_query(fd, -1, 0, AO64C_QUERY_NRATE_MIN, &nrate_min);

	fsamp_want	= (double) LIMIT_RANGE(fsamp, fsamp_min, fsamp_max);
	nrate_use	= fref / fsamp_want;
	nrate_ll	= nrate_use - 1;
	nrate_ul	= nrate_use + 1;
	nrate_ll	= (nrate_ll < nrate_min) ? nrate_min : nrate_ll;
	nrate_ul	= (nrate_ul > nrate_max) ? nrate_max : nrate_ul;

	for (nrate_use = nrate_ll; nrate_use <= nrate_ul; nrate_use++)
	{
		fsamp_got	= ((double) fref) / nrate_use;
		delta		= fsamp_got - fsamp_want;

		if ((ABS(delta)) < (ABS(delta_best)))
		{
			delta_best	= delta;
			fsamp_best	= fsamp_got;
			nrate_best	= nrate_use;
		}
	}

	nrate[0]	= nrate_best;

	if (rate)
		rate[0]	= fsamp_best;

	if (verbose)
	{
		printf("%s  (", errs ? " FAIL <---" : "");
		printf("Fref ");
		gsc_label_long_comma(fref);
		printf(", Nrate ");
		gsc_label_long_comma(nrate_best);
		printf(", Fsamp %.3f", fsamp_best);
		printf(")\n");
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	ao64c_fsamp_ao_get
*
*	Purpose:
*
*		Determine the sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	0: errors only, !0: all output
*
*		fsamp	Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao64c_fsamp_ao_get(int fd, int index, int verbose, s32* fsamp)
{
	int	errs		= 0;
	s32	fref;
	s32	nrate;
	s32	tmp;

	if (verbose)
		gsc_label_index("Sample Rate", index);

	if (fsamp == NULL)
		fsamp	= &tmp;

	errs	+= ao64c_query(fd, -1, 0, AO64C_QUERY_FREF_DEFAULT, &fref);
	errs	+= ao64c_rag_nrate(fd, -1, 0, -1, &nrate);

	fsamp[0]	= (s32) (((double) fref / nrate) + 0.5);

	if (verbose)
	{
		gsc_label_long_comma(fsamp[0]);
		printf(" S/S  (Fref ");
		gsc_label_long_comma(fref);
		printf(", Nrate = ");
		gsc_label_long_comma(nrate);
		printf(")\n");
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	ao64c_fsamp_ao_report
*
*	Purpose:
*
*		Determine and report the sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	0: errors only, !0: all output
*
*		chan	The index of the channel of interest.
*
*		fsamp	Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao64c_fsamp_ao_report(int fd, int index, int verbose, int chan, s32* fsamp)
{
	char	buf[64];
	s32		clk_mode;
	s32		clk_src;
	int		errs		= 0;
	s32		enable;
	s32		fref;
	s32		nrate;
	s32		tmp;

	if (verbose)
	{
		sprintf(buf, "Channel %d", chan);
		gsc_label_index(buf, index);
	}

	if (fsamp == NULL)
		fsamp	= &tmp;

	errs	+= ao64c_query		(fd, -1, 0, AO64C_QUERY_FREF_DEFAULT, &fref);
	errs	+= ao64c_rag_enable	(fd, -1, 0, -1, &enable);
	errs	+= ao64c_rag_nrate	(fd, -1, 0, -1, &nrate);
	errs	+= ao64c_clock_src	(fd, -1, 0, -1, &clk_src);
	errs	+= ao64c_clock_mode	(fd, -1, 0, -1, &clk_mode);

	fsamp[0]	= (s32) (((double) fref / nrate) + 0.5);

	if ((clk_src == AO64C_CLOCK_SRC_EXT) &&
		(clk_mode == AO64C_CLOCK_MODE_TARGET))
	{
		fsamp[0]	= 0;

		if (verbose)
			printf("0 S/S  (Clock Source: Cable Clock I/O)\n");
	}
	else if (enable == AO64C_RAG_ENABLE_NO)
	{
		fsamp[0]	= 0;

		if (verbose)
			printf("0 S/S  (Rate-A Generator Disabled)\n");
	}
	else
	{
		if (verbose)
		{
			gsc_label_long_comma(fsamp[0]);
			printf(" S/S  (Fref ");
			gsc_label_long_comma(fref);
			printf(", Nrate = ");
			gsc_label_long_comma(nrate);
			printf(")\n");
		}
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	ao64c_fsamp_ao_report_all
*
*	Purpose:
*
*		Determine and report the sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	0: errors only, !0: all output
*
*		map		The set of channels being used.
*
*		fsamp	Store the overall sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao64c_fsamp_ao_report_all(int fd, int index, int verbose, u64 map, s32* fsamp)
{
	u64	bit;
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("Sample Rates", index);

	errs	= ao64c_query(fd, -1, 0, AO64C_QUERY_CHANNEL_QTY, &chans);

	if (errs == 0)
	{
		if (verbose)
			printf("\n");

		gsc_label_level_inc();

		for (chan = 0; chan < chans; chan++)
		{
			bit	= (u64) 0x1 << chan;

			if ((bit & map) == 0)
				continue;

			sps		= 0;
			errs	+= ao64c_fsamp_ao_report(fd, -1, verbose, chan, &sps);
			total	+= sps;
		}

		if (fsamp)
			fsamp[0]	= total;

		if (verbose)
		{
			gsc_label("Overall Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");

		}

		gsc_label_level_dec();
	}

	return(errs);
}


